@@ -33,7 +33,7 @@
  * Changelog:
  *   0.0.6 Added support for different sample formats.
  *
- * speaker-test.c,v 1.26 2006/04/06 17:24:12 tiwai Exp
+ * $Id: speaker_test.c,v 1.00 2003/11/26 19:43:38 jcdutton Exp $
  */
 
 #include <stdio.h>
@@ -342,6 +342,8 @@ static int set_hwparams(snd_pcm_t *handle, snd_pcm_hw_params_t *params, snd_pcm_
   }
   if (! buffer_time && ! period_time) {
     buffer_size = buffer_size_max;
+    if (! period_time)
+      buffer_size = (buffer_size / nperiods) * nperiods;
     printf(_("Using max buffer size %lu\n"), buffer_size);
     err = snd_pcm_hw_params_set_buffer_size_near(handle, params, &buffer_size);
     if (err < 0) {
